@@ -1,4 +1,4 @@
-/* $Header: /cvsroot/macperl/perl/macos/ext/Mac/Carbon/Carbon.h,v 1.6 2002/12/17 16:18:37 pudge Exp $
+/* $Header: /cvsroot/macperl/perl/macos/ext/Mac/Carbon/Carbon.h,v 1.9 2003/04/07 01:25:31 pudge Exp $
  *
  *    Copyright (c) 2002 Matthias Neeracher, Chris Nandor
  *
@@ -6,6 +6,15 @@
  *    as specified in the README file.
  *
  * $Log: Carbon.h,v $
+ * Revision 1.9  2003/04/07 01:25:31  pudge
+ * Handle directory case
+ *
+ * Revision 1.8  2003/04/06 22:30:29  pudge
+ * Add GUSIFS2FullPath and GUSIPath2FS; do some cleanup
+ *
+ * Revision 1.7  2003/03/17 06:15:56  pudge
+ * A go at fixing the problems with the GUSI routines, so that if a path doesn't exist, we try to get the directory instead; still need some cleanup and a lot of testing
+ *
  * Revision 1.6  2002/12/17 16:18:37  pudge
  * Use new constant
  *
@@ -68,11 +77,14 @@ typedef Ptr	PtrRet;
 /* $! and $^E will both be set together, which is fine (for now?) */
 #define gMacPerl_OSErr (short)errno
 
+// keyReplyPortAttr is not found on some Mac OS X versions ...
+// so define our own.  nyah nyah.
+#define keyPerlReplyPortAttr 'repp'
+
 /* c string to pascal string */
 static StringPtr MacPerl_CopyC2P(const char * c, StringPtr p)
 {
 	memcpy(p+1, c, *p = strlen(c));
-	
 	return p;
 }
 
@@ -135,28 +147,110 @@ static UInt32 SecondsUnix2Mac(UInt32 unixseconds)
  *
  *   * Better error handling for odd cases
  *   * Alias resolving?  Is that even necessary?
- *   * Using FSRef routines apparently means it is harder to allow
- *     use of filenames that don't exist ... do we care?  Maybe for FindFolder() ...
+ *   * Can only handle 31 characters ... yes, that sucks
  *   * Probably bugs to fix ...
  *
  */
 
 
+/* This returns an FSSpec of the directory enclosing the given FSSpec */
+static OSErr GUSIFSpUp(FSSpec * spec)
+{
+	if (spec->parID == fsRtParID) {
+		spec->vRefNum = 666;
+		spec->parID   = 0;
+		spec->name[0] = 0;
+	} else {
+		if (1) { // CatInfo(-1)
+			gMacPerl_OSErr = FSMakeFSSpec(
+				spec->vRefNum, spec->parID, NULL, spec
+			);
+			if (gMacPerl_OSErr == fnfErr)
+				gMacPerl_OSErr = noErr;
+		}
+	}
+
+	return gMacPerl_OSErr;
+}
+
+/* This returns an FSSpec of the named file in the directory in the given FSSpec */
+static OSErr GUSIFSpDown(FSSpec * spec, ConstStr31Param name)
+{
+	CInfoPBRec *	pb      = malloc(sizeof(CInfoPBRec));
+	UInt8 *		path    = malloc(2*PATH_MAX); // to be safe
+
+	pb->dirInfo.ioVRefNum   = spec->vRefNum;
+	pb->dirInfo.ioDrDirID   = spec->parID;
+	pb->dirInfo.ioNamePtr   = spec->name;
+	pb->dirInfo.ioFDirIndex = 0;
+
+	gMacPerl_OSErr = PBGetCatInfoSync(pb);
+
+	if (!gMacPerl_OSErr) {
+		MacPerl_CopyC2P(name, path);
+		gMacPerl_OSErr = FSMakeFSSpec(
+			pb->dirInfo.ioVRefNum, pb->dirInfo.ioDrDirID, path, spec
+		);
+
+		// is directory, and no name given
+		if ((pb->dirInfo.ioFlAttrib & 0x10) && (*name == 0)) {
+			spec->parID = pb->dirInfo.ioDrDirID;
+			spec->name[0] = 0;
+		}
+
+		if (gMacPerl_OSErr == fnfErr)
+			gMacPerl_OSErr = noErr;
+	}
+
+	free(pb);
+	free(path);
+
+	return gMacPerl_OSErr;
+}
+
+
 /* Returns full POSIX-style path from FSSpec */
 /* Returns empty string on failure, sets gMacPerl_OSErr to nonzero */
 static char * GUSIFSp2FullPath(const FSSpec * spec)
 {
 	FSRef		ref;
-	UInt8 *		path     = (UInt8 *)NewPtr(2*PATH_MAX); // to be safe
+	char *		name     = malloc(256);
+	UInt8 *		path     = malloc(2*PATH_MAX); // to be safe
 	UInt32		pathSize = 2*PATH_MAX;
 
-	// File doesn't exist, big problem
-	if ( (gMacPerl_OSErr = FSpMakeFSRef(spec, &ref)) ) { // && (gMacPerl_OSErr != fnfErr) )
-		return "";
+	// If file doesn't exist, see if dir does
+	*name = 0;
+	if ( (gMacPerl_OSErr = FSpMakeFSRef(spec, &ref)) && (gMacPerl_OSErr == fnfErr) ) {
+		FSSpec *	spec2 = malloc(sizeof(FSSpec));
+
+		spec2->vRefNum = spec->vRefNum;
+		spec2->parID   = spec->parID;
+		spec2->name[0] = 0;
+
+		if ( GUSIFSpUp(spec2) ) {
+			free(name);
+			free(spec2);
+			return "";
+		}
+
+		if ( gMacPerl_OSErr = FSpMakeFSRef(spec2, &ref) ) {
+			free(name);
+			free(spec2);
+			return "";
+		} else {
+			strncpy(name, spec->name+1, *spec->name);
+			name[*spec->name] = 0;
+			free(spec2);
+		}
 	}
 
-	if (gMacPerl_OSErr = FSRefMakePath(&ref, path, pathSize))
-		return "";
+	if (gMacPerl_OSErr = FSRefMakePath(&ref, path, pathSize)) {
+		path = "";
+	} else if (*name != 0) {
+		sprintf(path, "%s/%s", path, name);
+	}
+
+	free(name);
 
 	return (char *)path;
 }
@@ -188,25 +282,65 @@ static OSErr GUSIPath2FSp(const char * fileName, FSSpec * spec)
  		switch (gMacPerl_OSErr = FSMakeFSSpec(spec->vRefNum, spec->parID, path, spec)) {
 		 	case fnfErr: // It is OK if file does not exist
 				gMacPerl_OSErr = noErr;
- 		 		return gMacPerl_OSErr;
+ 				return gMacPerl_OSErr;
 			default:
 				return gMacPerl_OSErr;
 		}
 	}
 
-	// File doesn't exist, big problem
-	if (gMacPerl_OSErr = FSPathMakeRef((UInt8 *)fileName, &ref, NULL))
-		return gMacPerl_OSErr;
+	// If file doesn't exist, see if dir does
+	if ( (gMacPerl_OSErr = FSPathMakeRef((UInt8 *)fileName, &ref, NULL)) && (gMacPerl_OSErr == fnfErr) ) {
+		char *	fileNameCpy = malloc(strlen(fileName));
+		char *	name;
+
+		strcpy(fileNameCpy, fileName);
+
+		if ( !(name = rindex(fileNameCpy, '/')) ) {
+			sprintf(fileNameCpy, "./%s", fileName);
+			name = rindex(fileNameCpy, '/');
+
+		} else if ( name && *name != 0 && name[1] == 0 ) {
+			fileNameCpy[name - fileNameCpy] = 0;
+			if (! (name = rindex(fileNameCpy, '/')) ) {
+				char * fileNameCpy2	= malloc(strlen(fileName));
+
+				strcpy(fileNameCpy2, fileNameCpy);
+				sprintf(fileNameCpy, "./%s", fileNameCpy2);
+				name = rindex(fileNameCpy, '/');
+
+				free(fileNameCpy2);
+			}
+		}
+
+		if ( name && *name != 0 ) {
+			char * dir		= malloc(PATH_MAX);
+
+			strncpy(dir, fileNameCpy, (name - fileNameCpy));
+			dir[name - fileNameCpy] = 0;
+			name++;
+			gMacPerl_OSErr = FSPathMakeRef((UInt8 *)dir, &ref, NULL);
+			free(dir);
+		}
+
+		if (gMacPerl_OSErr) {
+			free(fileNameCpy);
+			return gMacPerl_OSErr;
+		} else {
+			if (! (gMacPerl_OSErr = FSGetCatalogInfo(&ref, kFSCatInfoNone, NULL, NULL, spec, NULL)) )
+				GUSIFSpDown(spec, name);
+			free(fileNameCpy);
+		}
+	} else {
+		gMacPerl_OSErr = FSGetCatalogInfo(&ref, kFSCatInfoNone, NULL, NULL, spec, NULL);
+	}
 
-	// get FSSpec
-	gMacPerl_OSErr = FSGetCatalogInfo(&ref, kFSCatInfoNone, NULL, NULL, spec, NULL);
 	return gMacPerl_OSErr;
 }
 
 /* This returns the GUSI-specific text representation of an FSSpec */
 static char * GUSIFSp2Encoding(const FSSpec * spec)
 {
-	char * realSpec = NewPtr(255);
+	char * realSpec = malloc(255);
 	char fmt[32];
 
 	sprintf(fmt, "\021%%04hX%%08X:%%%d.%ds", *spec->name, *spec->name);
@@ -215,35 +349,6 @@ static char * GUSIFSp2Encoding(const FSSpec * spec)
 	return realSpec;
 }
 
-/* This returns an FSSpec of the directory enclosing the given FSSpec */
-static OSErr GUSIFSpUp(FSSpec * spec)
-{
-	if (spec->parID == fsRtParID) {
-		spec->vRefNum = 666;
-		spec->parID   = 0;
-		spec->name[0] = 0;
-	} else {
-		if (1) { // CatInfo(-1)
-			gMacPerl_OSErr = FSMakeFSSpec(
-				spec->vRefNum, spec->parID, NULL, spec
-			);
-		}
-	}
-
-	return gMacPerl_OSErr;
-}
-
-/* This returns an FSSpec of the named file in the directory in the given FSSpec */
-/* untested!! */
-static OSErr GUSIFSpDown(FSSpec * spec, ConstStr31Param name)
-{
-	gMacPerl_OSErr = FSMakeFSSpec(
-		spec->vRefNum, spec->parID, name, spec
-	);
-
-	return gMacPerl_OSErr;
-}
-
 /* Finds special folders */
 static OSErr GUSISpecial2FSp(OSType object, short vol, FSSpec * desc)
 {
@@ -253,12 +358,26 @@ static OSErr GUSISpecial2FSp(OSType object, short vol, FSSpec * desc)
 	return GUSIFSpUp(desc);
 }
 
+static char * GUSIFS2FullPath(const FSRef * ref)
+{
+	UInt8 *	path     = malloc(2*PATH_MAX); // to be safe
+	UInt32	pathSize = 2*PATH_MAX;
+
+	gMacPerl_OSErr = FSRefMakePath(ref, path, pathSize);
+	return path;
+}
+
+static OSErr GUSIPath2FS(const char * fileName, FSRef * ref)
+{
+	gMacPerl_OSErr = FSPathMakeRef((UInt8 *)fileName, ref, NULL);
+	return gMacPerl_OSErr;
+}
 
 
 static void fsetfileinfo(char * path, OSType creator, OSType type)
 {
 	FInfo info;
-	FSSpec * spec = (FSSpec *)NewPtr(sizeof(FSSpec));
+	FSSpec * spec = malloc(sizeof(FSSpec));
 
 	if ((gMacPerl_OSErr = GUSIPath2FSp(path, spec)) == 0) {
 		if ((gMacPerl_OSErr = FSpGetFInfo(spec, &info)) == 0) {
@@ -269,13 +388,14 @@ static void fsetfileinfo(char * path, OSType creator, OSType type)
 			gMacPerl_OSErr = FSpSetFInfo(spec, &info);
 		}
 	}
-	DisposePtr((char *)spec);
+
+	free(spec);
 }
 
 static void fgetfileinfo(char * path, OSType * creator, OSType * type)
 {
 	FInfo info;
-	FSSpec * spec = (FSSpec *)NewPtr(sizeof(FSSpec));
+	FSSpec * spec = malloc(sizeof(FSSpec));
 
 	if ((gMacPerl_OSErr = GUSIPath2FSp(path, spec)) == 0) {
 		if ((gMacPerl_OSErr = FSpGetFInfo(spec, &info)) == 0) {
@@ -285,34 +405,10 @@ static void fgetfileinfo(char * path, OSType * creator, OSType * type)
 				*type    = info.fdType;
 		}
 	}
-	DisposePtr((char *)spec);
+
+	free(spec);
 }
 
-// keyReplyPortAttr is not found on some Mac OS X versions ...
-// so define our own.  nyah nyah.
-#define keyPerlReplyPortAttr 'repp'
-
-/* Maybe we'll have use for this again? *
-* 
-* static pascal Boolean SubLaunchIdle(EventRecord * ev, long * sleep, RgnHandle * rgn)
-* {
-* 	// printf("%d : %d : %d : %d\n", kHighLevelEvent, ev->what, sleep, rgn);
-* 	switch (ev->what) {
-* 		case kHighLevelEvent:
-* 			AEProcessAppleEvent(ev);
-* 			return true;
-* 		default:
-* 			break;
-* 	}		
-* 
-* 	*sleep = 0;
-* 	rgn = nil;
-* 	return false;
-* }
-*/
-
-//#define uSubLaunchIdle *(AEIdleUPP)&SubLaunchIdle
-
 #  endif /* MACOS_TRADITIONAL */
 
 #endif /* _MAC_CARBON_H */
